/*
 * #%L
 * BigDataViewer core classes with minimal dependencies
 * %%
 * Copyright (C) 2012 - 2016 Tobias Pietzsch, Stephan Saalfeld, Stephan Preibisch,
 * Jean-Yves Tinevez, HongKee Moon, Johannes Schindelin, Curtis Rueden, John Bogovic
 * %%
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * #L%
 */
package bdv.fx.viewer.render;

import java.util.concurrent.ExecutorService;

import bdv.cache.CacheControl;
import bdv.viewer.render.AccumulateProjectorFactory;
import net.imglib2.Interval;
import net.imglib2.Volatile;
import net.imglib2.img.array.ArrayImg;
import net.imglib2.type.numeric.ARGBType;
import net.imglib2.ui.RenderTarget;
import net.imglib2.ui.Renderer;

/**
 * A {@link Renderer} that uses a coarse-to-fine rendering scheme. First, a small {@link ArrayImg} at a fraction of
 * the canvas resolution is rendered. Then, increasingly larger images are rendered, until the full canvas resolution is
 * reached.
 * <p>
 * When drawing the low-resolution {@link ArrayImg} to the screen, they will be scaled up by Java2D to the full canvas
 * size, which is relatively fast. Rendering the small, low-resolution images is usually very fast, such that the
 * display is very interactive while the user changes the viewing transformation for example. When the transformation
 * remains fixed for a longer period, higher-resolution details are filled in successively.
 * <p>
 * The renderer allocates a {@link ArrayImg} for each of a predefined set of
 * <em>screen scales</em> (a screen scale of 1 means that 1 pixel in the screen
 * image is displayed as 1 pixel on the canvas, a screen scale of 0.5 means 1 pixel in the screen image is displayed as
 * 2 pixel on the canvas, etc.)
 * <p>
 * At any time, one of these screen scales is selected as the <em>highest screen scale</em>. Rendering starts with this
 * highest screen scale and then proceeds to lower screen scales (higher resolution images). Unless the highest screen
 * scale is currently rendering, {@link #requestRepaint(Interval) repaint request} will cancel rendering, such that display
 * remains interactive.
 * <p>
 * The renderer tries to maintain a per-frame rendering time close to a desired number of <code>targetRenderNanos</code>
 * nanoseconds. If the rendering time (in nanoseconds) for the (currently) highest scaled screen image is above this
 * threshold, a coarser screen scale is chosen as the highest screen scale to use. Similarly, if the rendering time for
 * the (currently) second-highest scaled screen image is below this threshold, this finer screen scale chosen as the
 * highest screen scale to use.
 * <p>
 * The renderer uses multiple threads (if desired) and double-buffering (if desired).
 * <p>
 * Double buffering means that three {@link ArrayImg BufferedImages} are created for every screen scale. After
 * rendering the first one of them and setting it to the {@link RenderTarget}, next time, rendering goes to the second
 * one, then to the third. The {@link RenderTarget} will always have a complete image, which is not rendered to while it
 * is potentially drawn to the screen. When setting an image to the {@link RenderTarget}, the {@link RenderTarget} will
 * release one of the previously set images to be rendered again. Thus, rendering will not interfere with painting the
 * {@link ArrayImg} to the canvas.
 * <p>
 * The renderer supports rendering of {@link Volatile} sources. In each rendering pass, all currently valid data for the
 * best fitting mipmap level and all coarser levels is rendered to a {@link #renderImages temporary image} for each
 * visible source. Then the temporary images are combined to the final image for display. The number of passes required
 * until all data is valid might differ between visible sources.
 * <p>
 * Rendering timing is tied to a {@link CacheControl} control for IO budgeting, etc.
 *
 * @author Tobias Pietzsch &lt;tobias.pietzsch@gmail.com&gt;
 */
public class MultiResolutionRendererFX extends MultiResolutionRendererGeneric<BufferExposingWritableImage>
{

	public static class MakeWritableImage
			implements MultiResolutionRendererGeneric.ImageGenerator<BufferExposingWritableImage>
	{

		@Override
		public BufferExposingWritableImage create(final int width, final int height)
		{
			try
			{
				return new BufferExposingWritableImage(width, height);
			} catch (final Exception e)
			{
				throw e instanceof RuntimeException ? (RuntimeException) e : new RuntimeException(e);
			}
		}

		@Override
		public BufferExposingWritableImage create(final int width, final int height, final BufferExposingWritableImage
				other)
		{
			// TODO can we somehow re-use smaller image?
			return create(width, height);
		}

	}

	public MultiResolutionRendererFX(
			final TransformAwareRenderTargetGeneric<BufferExposingWritableImage> display,
			final PainterThread painterThread,
			final double[] screenScales,
			final long targetRenderNanos,
			final boolean doubleBuffered,
			final int numRenderingThreads,
			final ExecutorService renderingExecutorService,
			final boolean useVolatileIfAvailable,
			final AccumulateProjectorFactory<ARGBType> accumulateProjectorFactory,
			final CacheControl cacheControl)
	{
		super(
				display,
				painterThread,
				screenScales,
				targetRenderNanos,
				doubleBuffered,
				numRenderingThreads,
				renderingExecutorService,
				useVolatileIfAvailable,
				accumulateProjectorFactory,
				cacheControl,
				BufferExposingWritableImage::asArrayImg,
				new MakeWritableImage(),
				img -> (int) img.getWidth(),
				img -> (int) img.getHeight()
		     );
	}

}
